/*
 #
 # Filename: volume_sm3_parallel.fp
 #
 # Author: Wu Shin-Ting (26/12/2010)
 # Modified: Lionis de Souza Watanabe
 */

//#extension GL_EXT_gpu_shader4 : enable //necessary for bitwise operations

uniform sampler1D TRANSFERFUNCTION;
uniform sampler3D IMAGE3D;
uniform sampler3D CONTROLVOLUME;
uniform vec3 cursorPosition;
uniform vec4 scaleFactor;
uniform vec3 extents;
uniform float stepSize;
uniform bvec3 drawPlaneSide1;
uniform bvec3 drawPlaneSide2;
uniform bvec3 clippingPlaneSide1;
uniform bvec3 clippingPlaneSide2;
uniform float depth_slicing;
uniform float noise_threshold;

float z_value;

// Find the ray intersection interval with respect 
// to the proxy geometry  (Liang Barsky)
bool intersectBox(in vec4 r_o, in vec4 r_d, in vec4 boxmin, in vec4 boxmax, 
inout float tnear, inout float tfar)
{
    // compute intersection of ray with all six bbox planes (Liang-Barsky)
    vec4 invR = vec4(1.0,1.0,1.0,1.0)/r_d;
    vec4 tbot = invR * (boxmin - r_o);
    vec4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    vec4 tmin = min(ttop, tbot);
    vec4 tmax = max(ttop, tbot);

    if (all(notEqual (tmin.xyz,tmax.xyz))) {
        // find the largest tmin and the smallest tmax
    	float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    	float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

    	tnear = largest_tmin;
    	tfar = smallest_tmax;

    	if (tfar >= tnear) return true;
    	else return false;
    } else {
      // r_d is parallel to at least two planes. Just consider the range [0,1].
      // It should be improved.
      tnear = 0.;
      tfar = 1.0;
      return true;
    }
}

//Process geometric position from any voxel,
//copying its depth to depth buffer (gl_FragDepth).
//thus any voxels between a observer and this voxel
//will be ignored from visualization
void processZvalue(in vec4 geomPos, inout bool data_found){
	vec4 centralizedGeomPos, boundary;
	mat4 mvp;
	
	mvp = gl_ModelViewProjectionMatrix;
	boundary.xyz = extents.xyz;
	boundary.w = 1.0;

	centralizedGeomPos = boundary * geomPos; // geomPos range: [0,1]; centralizedGeomPos range: [-1,1]
	z_value = (mvp * centralizedGeomPos).z;// transform the centralized geomPos
	z_value = 0.5 * z_value + 0.5;//put it in range [0,1]
	
	data_found = true;
}

//Process all control tags (markups and depths, at this moment)
void processControlTags(in float image3Dscalar, in float TFscalar, 
in vec4 geomPos, inout bool data_found, inout vec4 dstColor){
	int markup;
	float depth;
	vec4 controlTag;

	controlTag = texture3D(CONTROLVOLUME, geomPos.xyz);
	markup = int(controlTag.x * float(INT_FORMAT_BITS));
	
	//all markups (except transparency) are considered valid voxels
	//to be included in depth buffer
	switch (markup){
	case TAG001://transparent
		break;
	case TAG002://painted
		dstColor = vec4(0.0, 1.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG003://point of reference
		dstColor = vec4(1.0, 0.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG004://point of interest
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG005://line
		dstColor = vec4(0.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG006://eraser
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	default: //if there wasn't a markup tag, begins depth processing 
		depth = controlTag.y;//get voxel's depth //Revised by Ting
		
		//zvalue is set once, only to the first valid voxel in ray composition.
		//so, if some valid voxel was already found, depth buffer from previous voxel
		//won't be ovewritten, ignoring this step
		if(!data_found)
		
			//1) all untagged voxels have depth=0 and must be shown.
			//2) also, voxels deeper (with greater depth) than the slicing mesh have to be shown.
			if(depth == 0. || depth_slicing <= depth)
			
				//a scalar greater than noise threshold determines a valid voxel
				if(image3Dscalar > noise_threshold)
				
					//TFscalar is what the user is seeing indeed,
					//so consider only valid entries from transfer function
					if(TFscalar > 0.0)
						processZvalue(geomPos, data_found);
		break;
	}
}

//Process the axis aligned planes, showing them and
//doing volume clipping if desired by the user
void processAxisPlanes(in vec4 geomPos, in vec4 axialColor, 
in vec4 sagittalColor, in vec4 coronalColor, inout bool data_found, 
inout vec4 dstColor){
	//Drawing planes from side 1
	if (drawPlaneSide1[vmtkAXIAL] == true && geomPos.y > cursorPosition.z
			&& geomPos.y < cursorPosition.z + stepSize) {
		dstColor += axialColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[vmtkSAGITTAL] == true && geomPos.z > cursorPosition.x
			&& geomPos.z < cursorPosition.x + stepSize) {
		dstColor += sagittalColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[vmtkCORONAL] == true && geomPos.x > cursorPosition.y
			&& geomPos.x < cursorPosition.y + stepSize) {
		dstColor += coronalColor;
		processZvalue(geomPos, data_found);
	}

	//Clipping planes from side 1
	if ((clippingPlaneSide1[vmtkAXIAL] == true && geomPos.y < cursorPosition.z)
		|| (clippingPlaneSide1[vmtkSAGITTAL] == true && geomPos.z < cursorPosition.x)
		|| (clippingPlaneSide1[vmtkCORONAL] == true && geomPos.x < cursorPosition.y)){

		//no color until ray is passing through clipped portion of volume
		dstColor = vec4(0.0);
		
		z_value = 1.0;

		//the clipped portion of volume isn't valid data to be 
		//included in ray composition, so no data found until this position
		data_found = false;
	}
}

void main(void) {
	int i;
	bool data_found, hit;
	float scalar, t, tnear, tfar;
	vec4 geomPos, initGeomPos, volExtentMin, volExtentMax;
	vec4 axialColor, sagittalColor, coronalColor;
	vec4 dstColor, srcColor, camDir;
	vec4 geomDir;
	mat4 mvi;

	axialColor = vec4(0.0, 0.08, 0.08, 0.2);//color of axial plane
	sagittalColor = vec4(0.08, 0.0, 0.08, 0.2);//color of sagittal plane
	coronalColor = vec4(0.08, 0.08, 0.0, 0.2);//color of coronal plane

	volExtentMin = vec4(0.0, 0.0, 0.0, 0.0);//min value of boundingbox
	volExtentMax = vec4(1.0, 1.0, 1.0, 1.0);//max value of boundingbox

	dstColor = srcColor = vec4(0.0);//initialize color
	//gl_FragDepth = 1.0;//set default depth to 1.0
	z_value = 1.0;	

	data_found = false; //flag used when valid voxels are found
	camDir = vec4(0.0, 0.0, -1.0, 0.0);//camera direction, used for ray composition

	mvi = gl_ModelViewMatrixInverse;//inverse model view matrix

	//determine volume entry position / ray's starting point
	initGeomPos = gl_TexCoord[1];
	initGeomPos.w = 1.0;

	//compute ray direction
	geomDir = normalize(mvi * camDir) * scaleFactor;
//	geomDir = mvi * camDir;
	geomDir.w = 0.0;

	hit = intersectBox(initGeomPos, geomDir, volExtentMin, volExtentMax, tnear, tfar);
    if (!hit) {
       gl_FragColor = vec4(1.0,0.0,0.0,1.0);
       gl_FragDepth = z_value;
       return;
    }

    if (tnear < 0.0) tnear = 0.0;     // clamp to near plane
	t = tnear;
	///the ray traversal loop checks if the ray is still inside of volume extents
	//and if opacity haven't reached maximum value (opacity = 1)
	while (t < tfar && dstColor.a < 0.99) {
		geomPos = initGeomPos + geomDir * t;
		scalar = texture3D(IMAGE3D, geomPos.xyz).r;//get RAW scalar value
		srcColor = texture1D(TRANSFERFUNCTION, scalar);//apply transfer function
		srcColor *= srcColor.a; //compute associated color of voxel
		
		//control volume processing determines depth buffer and
		//the beginning of ray composition
		processControlTags(scalar, srcColor.a, geomPos, data_found, dstColor);

		//if a valid voxel was found, begin/continue front-to-back ray composition
		if(data_found){				
			//linear interpolation of dstColor and srcColor
			//opacity is just accumulated
			dstColor.rgb += mix(vec3(0.0), srcColor.rgb, 1.0 - dstColor.a);
			dstColor.a += srcColor.a;
		}
		
		//draw or clip the volume
		processAxisPlanes(geomPos, axialColor, sagittalColor, coronalColor, data_found, dstColor);
					
		//advance ray position along ray direction
		t += stepSize;
	}
	
	gl_FragColor = dstColor;
	gl_FragDepth = z_value;
	//gl_FragColor = vec4(z_value,z_value,z_value,0.5);
}
